/*********************************************************************
*
*       SFDTD: Finite Difference Time Domain analysis program
*		for Wire Antenna Version 1.0
*
*	make: Linux 2.0.30
*               gcc -lm sfdtd.c
*
*       Jan 16, 1999
*       Shonan Radio Laboratory
*       Mutsumi Shimazaki (JR1KNA)	mail:shima@labnet.or.jp
*
*********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define XMAX 120
#define YMAX 120
#define ZMAX 40
#define MAXCHAR 255
#define MAXCOND 10
#define INPFILE "input.dat"
#define OUTFILE "out"

int getword(char *word_g,int number_g,char *line_g);
int inputdata(double *dx,double *dy,double *dz,
	int *maxstep,int *outstep,int *outplane,
	int *extx,int *exty,int *extz,int *extd,
	double *extf,double *exta,int *cons,
	int *cond,int *con1,int *con2,int *con3,int *con4);

main()
{
    int i,j,k;
    int nstep,maxstep,outstep,outplane,fnum=0;
    char filenamex[MAXCHAR],filenamey[MAXCHAR],filenamez[MAXCHAR];
    double dex,dey,dez,dhx,dhy,dhz;
    double c,pai,myuo,epuo,epu,myu;
    double dt,dx,dy,dz;
    int extx,exty,extz,extd;
    double extf,exta,val;
    int cons,cond[MAXCOND];
    int con1[MAXCOND],con2[MAXCOND],con3[MAXCOND],con4[MAXCOND];
    FILE *rfp,*wfpx,*wfpy,*wfpz;

    double *ex0=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *ey0=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *ez0=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *ex1=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *ey1=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *ez1=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *hx0=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *hy0=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *hz0=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *hx1=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *hy1=malloc(XMAX*YMAX*ZMAX*sizeof(double));
    double *hz1=malloc(XMAX*YMAX*ZMAX*sizeof(double));

    printf("********** SFDTD PROGRAM **********\n");
    printf("Initialize Configration:\n");
    if(inputdata(&dx,&dy,&dz,&maxstep,&outstep,&outplane,
	    &extx,&exty,&extz,&extd,&extf,&exta,
	    &cons,cond,con1,con2,con3,con4)!=6){
	printf("\tInput data error.\n");
	exit(1);
    }

    c=2.997925e+8;
    pai=4.0*atan(1.0);
    myuo=pai*4.0e-7;
    epuo=1.0/(myuo*c*c);
    dt=1/(c*sqrt((1/(dx*dx))+(1/(dy*dy))+(1/(dz*dz))));
    epu=dt/epuo;
    myu=dt/myuo;

    printf("\tMesh \tsize \tx=%d y=%d z=%d\n",XMAX,YMAX,ZMAX);
    printf("\t\tdelta \tdx=%e dy=%e dz=%e\n",dx,dy,dz);
    printf("\tStep \tmax=%d output=%d \tdt=%e\n",maxstep,outstep,dt);
    printf("\tOutput \tplane=%d (1=xy,2=yz,3=zx)\n",outplane);
    printf("\tFeed \tx=%d y=%d z=%d \tdirection=%d (1=x,2=y,3=z)\n",
	extx,exty,extz,extd);
    printf("\tSource \tfrequency=%e amplitude=%e\n",extf,exta);
    printf("\tConductor \tnumbers=%d\n",cons);
    printf("Start Analysis:\n");

//  *** CLEAR VALUE ***
    for(i=0;i<XMAX*YMAX*ZMAX-1;i++){
	ex0[i]=0.0;
	ey0[i]=0.0;
	ez0[i]=0.0;
	ex1[i]=0.0;
	ey1[i]=0.0;
	ez1[i]=0.0;
	hx0[i]=0.0;
	hy0[i]=0.0;
	hz0[i]=0.0;
	hx1[i]=0.0;
	hy1[i]=0.0;
	hz1[i]=0.0;
    }

//  *** ITERATION LOOP ***
    printf("step \ttime \t\tsource\n");
    for(nstep=0;nstep<maxstep+1;nstep++){
	val=exta*sin(2*pai*extf*nstep*dt);
	switch(extd){
	    case 1:	ex0[(extx)*YMAX*ZMAX+(exty)*ZMAX+(extz)]=val;
			break;
	    case 2:	ey0[(extx)*YMAX*ZMAX+(exty)*ZMAX+(extz)]=val;
			break;
	    case 3:	ez0[(extx)*YMAX*ZMAX+(exty)*ZMAX+(extz)]=val;
			break;
	    default:	;
	}
	printf("%d \t%e \t%e\n",nstep,nstep*dt,val);

//	*** CALCULATE MAGNETIC FIELD ***
	for(i=1;i<XMAX-1;i++){
	    for(j=1;j<YMAX-1;j++){
		for(k=1;k<ZMAX-1;k++){
		    dey=(ey0[(i)*YMAX*ZMAX+(j)*ZMAX+(k+1)]
			-ey0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)])/dz;
		    dez=(ez0[(i)*YMAX*ZMAX+(j+1)*ZMAX+(k)]
			-ez0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)])/dy;
		    hx1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]=
			hx0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]+myu*(dey-dez);
		}
	    }
	}
	for(i=1;i<XMAX-1;i++){
	    for(j=1;j<YMAX-1;j++){
		for(k=1;k<ZMAX-1;k++){
		    dez=(ez0[(i+1)*YMAX*ZMAX+(j)*ZMAX+(k)]
			-ez0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)])/dx;
		    dex=(ex0[(i)*YMAX*ZMAX+(j)*ZMAX+(k+1)]
			-ex0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)])/dz;
		    hy1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]=
			hy0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]+myu*(dez-dex);
		}
	    }
	}
	for(i=1;i<XMAX-1;i++){
	    for(j=1;j<YMAX-1;j++){
		for(k=1;k<ZMAX-1;k++){
		    dex=(ex0[(i)*YMAX*ZMAX+(j+1)*ZMAX+(k)]
			-ex0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)])/dy;
		    dey=(ey0[(i+1)*YMAX*ZMAX+(j)*ZMAX+(k)]
			-ey0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)])/dx;
		    hz1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]=
			hz0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]+myu*(dex-dey);
		}
	    }
	}

//	*** CALCULATE ELECTRIC FIELD ***
	for(i=1;i<XMAX-1;i++){
	    for(j=1;j<YMAX-1;j++){
		for(k=1;k<ZMAX-1;k++){
		    dhy=(hy1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]
			-hy1[(i-1)*YMAX*ZMAX+(j)*ZMAX+(k)])/dx;
		    dhx=(hx1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]
			-hx1[(i)*YMAX*ZMAX+(j-1)*ZMAX+(k)])/dy;
		    ez1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]=
			ez0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]+epu*(dhy-dhx);
		}
	    }
	}
	for(i=1;i<XMAX-1;i++){
	    for(j=1;j<YMAX-1;j++){
		for(k=1;k<ZMAX-1;k++){
		    dhx=(hx1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]
			-hx1[(i)*YMAX*ZMAX+(j)*ZMAX+(k-1)])/dz;
		    dhz=(hz1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]
			-hz1[(i-1)*YMAX*ZMAX+(j)*ZMAX+(k)])/dx;
		    ey1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]=
			ey0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]+epu*(dhx-dhz);
		}
	    }
	}
	for(i=1;i<XMAX-1;i++){
	    for(j=1;j<YMAX-1;j++){
		for(k=1;k<ZMAX-1;k++){
		    dhz=(hz1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]
			-hz1[(i)*YMAX*ZMAX+(j-1)*ZMAX+(k)])/dy;
		    dhy=(hy1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]
			-hy1[(i)*YMAX*ZMAX+(j)*ZMAX+(k-1)])/dz;
		    ex1[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]=
			ex0[(i)*YMAX*ZMAX+(j)*ZMAX+(k)]+epu*(dhz-dhy);
		}
	    }
	}

//	*** CONDUCTOR ***
	for(i=0;i<cons;i++){
	    switch(cond[i]){
		case 1:	for(j=con1[i];j<=con2[i];j++)
			    ex1[(j)*YMAX*ZMAX+(con3[i])*ZMAX+(con4[i])]=0.0;
			break;
		case 2:	for(j=con2[i];j<=con3[i];j++)
			    ey1[(con1[i])*YMAX*ZMAX+(j)*ZMAX+(con4[i])]=0.0;
			break;
		case 3:	for(j=con3[i];j<=con4[i];j++)
			    ez1[(con1[i])*YMAX*ZMAX+(con2[i])*ZMAX+(j)]=0.0;
			break;
		default:;
	    }
	}

//	*** FILE OUTPUT ***
        if(nstep==outstep*fnum){
	    sprintf(filenamex,"%sx%d",OUTFILE,fnum);
	    sprintf(filenamey,"%sy%d",OUTFILE,fnum);
	    sprintf(filenamez,"%sz%d",OUTFILE,fnum);
	    wfpx=fopen(filenamex,"w");
	    wfpy=fopen(filenamey,"w");
	    wfpz=fopen(filenamez,"w");
	    switch(outplane){
		case 1: for(i=0;i<XMAX;i++){
	        	    for(j=0;j<YMAX;j++){
		    		fprintf(wfpx,"%d \t%d \t%e\n"
				    ,i,j,ex1[(i)*YMAX*ZMAX+(j)*ZMAX+(extz)]);
		    		fprintf(wfpy,"%d \t%d \t%e\n"
				    ,i,j,ey1[(i)*YMAX*ZMAX+(j)*ZMAX+(extz)]);
		    		fprintf(wfpz,"%d \t%d \t%e\n"
				    ,i,j,ez1[(i)*YMAX*ZMAX+(j)*ZMAX+(extz)]);
	        	    }
	    		}
			break;
		case 2: for(j=0;j<YMAX;j++){
	        	    for(k=0;k<ZMAX;k++){
		    		fprintf(wfpx,"%d \t%d \t%e\n"
				    ,j,k,ex1[(extx)*YMAX*ZMAX+(j)*ZMAX+(k)]);
		    		fprintf(wfpy,"%d \t%d \t%e\n"
				    ,j,k,ey1[(extx)*YMAX*ZMAX+(j)*ZMAX+(k)]);
		    		fprintf(wfpz,"%d \t%d \t%e\n"
				    ,j,k,ez1[(extx)*YMAX*ZMAX+(j)*ZMAX+(k)]);
	        	    }
	    		}
			break;
		case 3: for(i=0;i<XMAX;i++){
	        	    for(k=0;k<ZMAX;k++){
		    		fprintf(wfpx,"%d \t%d \t%e\n"
				    ,i,k,ex1[(i)*YMAX*ZMAX+(exty)*ZMAX+(k)]);
		    		fprintf(wfpy,"%d \t%d \t%e\n"
				    ,i,k,ey1[(i)*YMAX*ZMAX+(exty)*ZMAX+(k)]);
		    		fprintf(wfpz,"%d \t%d \t%e\n"
				    ,i,k,ez1[(i)*YMAX*ZMAX+(exty)*ZMAX+(k)]);
	        	    }
	    		}
			break;
	    }
	    fclose(wfpx);
	    fclose(wfpy);
	    fclose(wfpz);
	    fnum++;
        }

//	*** SET PREVIOUS VALUE ***
        for(i=1;i<XMAX*YMAX*ZMAX-1;i++){
	    ex0[i]=ex1[i];
	    ey0[i]=ey1[i];
	    ez0[i]=ez1[i];
	    hx0[i]=hx1[i];
	    hy0[i]=hy1[i];
	    hz0[i]=hz1[i];
        }
    }
    
    free(ex0);
    free(ey0);
    free(ez0);
    free(ex1);
    free(ey1);
    free(ez1);
    free(hx0);
    free(hy0);
    free(hz0);
    free(hx1);
    free(hy1);
    free(hz1);

}

int inputdata(double *dx,double *dy,double *dz,
	int *maxstep,int *outstep,int *outplane,
	int *extx,int *exty,int *extz,int *extd,
	double *extf,double *exta,int *cons,
	int *cond,int *con1,int *con2,int *con3,int *con4)
{
    FILE *rfp;
    int i,k=0;
    char line[MAXCHAR],word[MAXCHAR];

    rfp=fopen(INPFILE,"r");
    while(fgets(line,MAXCHAR,rfp)!=NULL){
        if(getword(word,1,line)==4){
            if(!strcmp(word,"DELTA")){
                getword(word,2,line);
                *dx=atof(word);
                getword(word,3,line);
                *dy=atof(word);
                getword(word,4,line);
                *dz=atof(word);
		k++;
            }
        }
	if(getword(word,1,line)==3){
            if(!strcmp(word,"STEP")){
                getword(word,2,line);
                *maxstep=atoi(word);
                getword(word,3,line);
                *outstep=atoi(word);
		k++;
	    }
        }
	if(getword(word,1,line)==2){
            if(!strcmp(word,"OUTPUT")){
                getword(word,2,line);
                *outplane=atoi(word);
		k++;
	    }
        }
	if(getword(word,1,line)==5){
            if(!strcmp(word,"FEED")){
                getword(word,2,line);
                *extx=atoi(word);
                getword(word,3,line);
                *exty=atoi(word);
                getword(word,4,line);
                *extz=atoi(word);
                getword(word,5,line);
                *extd=atoi(word);
		k++;
	    }
        }
	if(getword(word,1,line)==3){
            if(!strcmp(word,"SOURCE")){
                getword(word,2,line);
                *extf=atof(word);
                getword(word,3,line);
                *exta=atof(word);
		k++;
	    }
        }
	if(getword(word,1,line)==2){
            if(!strcmp(word,"CONDUCTORS")){
                getword(word,2,line);
                *cons=atoi(word);
		for(i=0;i<*cons;i++){
    		    if(fgets(line,MAXCHAR,rfp)!=NULL){
			if(getword(word,1,line)==6){
            		    if(atoi(word)==i+1){
				getword(word,2,line);
				cond[i]=atoi(word);
				getword(word,3,line);
				con1[i]=atoi(word);
				getword(word,4,line);
				con2[i]=atoi(word);
				getword(word,5,line);
				con3[i]=atoi(word);
				getword(word,6,line);
				con4[i]=atoi(word);
			    }
			}
		    }
		}
		k++;
	    }
	}
    }
    fclose(rfp);
    return k;
}

int getword(char *word_g,int number_g,char *line_g)
{
    int word_num=1;

    while(*line_g!='\0'){
        if(isgraph(*line_g)){
            if(word_num==number_g){
                *word_g=*line_g;
                word_g++;
            }
            line_g++;
            if(!isgraph(*line_g))
                word_num++;
        }
        else
            line_g++;
    }
    *word_g='\0';
    return word_num-1;
}
